from numpy.typing import NDArray
from typing import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.uniform import Uniform

import numpy as np
from ts3l.models.common import MLP


class SCARF(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dim,
        output_dim,
        encoder_depth=4,
        head_depth=2,
        dropout_rate = 0.04,
    ):
        """Implementation of SCARF: Self-Supervised Contrastive Learning using Random Feature Corruption.
        It consists in an encoder that learns the embeddings.
        It is done by minimizing the contrastive loss of a sample and a corrupted view of it.
        The corrupted view is built by replacing a random set of features by another sample randomly drawn independently.
            Args:
                input_dim (int): The size of the inputs.
                hidden_dim (int): The dimension of the hidden layers.
                output_dim (int): The dimension of output.
                encoder_depth (int, optional): The number of layers of the encoder MLP. Defaults to 4.
                head_depth (int, optional): The number of layers of the pretraining head. Defaults to 2.
                dropout_rate (float, optional): A hyperparameter that is to control dropout layer. Default is 0.04.
        """
        super().__init__()

        self.encoder = MLP(input_dim, hidden_dim, encoder_depth)

        self.pretraining_head = MLP(hidden_dim, hidden_dim, head_depth)

        self.head = nn.Sequential(
            OrderedDict([
                ("head_activation", nn.ReLU(inplace=True)),
                ("head_batchnorm", nn.BatchNorm1d(hidden_dim)),
                ("head_dropout", nn.Dropout(dropout_rate)),
                ("head_linear", nn.Linear(hidden_dim, output_dim))
            ])
        )

            
    def set_first_phase(self):
        self.forward = self.__first_phase_step
    
    def set_second_phase(self):
        self.forward = self.__second_phase_step

    def __first_phase_step(self, x, x_corrupted):

        emb_anchor = self.encoder(x)
        emb_anchor = self.pretraining_head(emb_anchor)

        emb_anchor = F.normalize(emb_anchor, p=2)
        
        emb_corrupted = self.encoder(x_corrupted)
        emb_corrupted = self.pretraining_head(emb_corrupted)
        emb_corrupted = F.normalize(emb_corrupted, p=2)

        return emb_anchor, emb_corrupted
    
    def __second_phase_step(self, x):
        emb = self.encoder(x)
        output = self.head(emb)

        return output



import torch
from torch import nn
from typing import OrderedDict

class MLP(nn.Sequential):
    """Simple multi-layer perceptron with ReLu activation and optional dropout layer"""

    def __init__(self, input_dim, hidden_dim, n_layers, dropout=0.0):
        layers = []
        in_dim = input_dim
        
        for i in range(n_layers - 1):
            layers.append(("linear_%d" % i, torch.nn.Linear(in_dim, hidden_dim)))
            layers.append(("batchnorm_%d" % i, nn.BatchNorm1d(hidden_dim)))
            layers.append(("relu_%d" % i, nn.ReLU(inplace=True)))
            layers.append(("dropout_%d" % i, torch.nn.Dropout(dropout)))
            in_dim = hidden_dim

        layers.append(("linear_n_layers", torch.nn.Linear(in_dim, hidden_dim)))

        super().__init__(OrderedDict(layers))


import torch
import torch.nn as nn

class VIMESelfSupervised(nn.Module):
    def __init__(self, input_dim:int):
        """Initialize self-supervised module of VIME

        Args:
            input_dim (int): The dimension of the encoder
        """
        super().__init__()
        self.h = nn.Linear(input_dim, input_dim, bias=True)
        self.mask_output = nn.Linear(input_dim, input_dim, bias=True)
        self.feature_output = nn.Linear(input_dim, input_dim, bias=True)

    def forward(self, x):
        """The forward pass of self-supervised module of VIME

        Args:
            x (torch.FloatTensor): The input batch.

        Returns:
            torch.FloatTensor: The predicted mask vector of VIME
            torch.FloatTensor: The predicted features of VIME
        """
        h = torch.relu(self.h(x))
        mask = torch.sigmoid(self.mask_output(h))
        feature = torch.sigmoid(self.feature_output(h))
        return mask, feature



from misc import get_args

import sys
import os
here = os.path.dirname(__file__)

sys.path.append(os.path.join(here, '..'))

def test_switchtab_classification():
    
    from benchmark.datasets import load_diabetes
    data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams = load_diabetes()
    
    sys.path.append(os.path.join(os.path.join(here, '..'), "benchmark"))
    
    from benchmark.pipelines import SwitchTabPipeLine
    
    args = get_args()
    args.data = "diabetes"
    
    pipeline = SwitchTabPipeLine(args, data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams)
    
    pipeline.benchmark()

def test_switchtab_regression():
    
    from benchmark.datasets import load_abalone
    data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams = load_abalone()
    
    sys.path.append(os.path.join(os.path.join(here, '..'), "benchmark"))
    
    from benchmark.pipelines import SwitchTabPipeLine
    
    args = get_args()
    args.data = "abalone"
    
    pipeline = SwitchTabPipeLine(args, data, label, continuous_cols, category_cols, output_dim, metric, metric_hparams)
    
    pipeline.benchmark()
    
if __name__ == "__main__":
    test_switchtab_classification()
    test_switchtab_regression()


from typing import Dict, Any, Type
import torch

from .base_module import TS3LLightining
from ts3l.models import SCARF
from ts3l.utils.scarf_utils import NTXentLoss, SCARFConfig

class SCARFLightning(TS3LLightining):
    
    def __init__(self, config: SCARFConfig) -> None:
        """Initialize the pytorch lightining module of SCARF

        Args:
            config (SubTabConfig): The configuration of SCARFLightning.
        """
        super(SCARFLightning, self).__init__(config)

    def _initialize(self, config: Dict[str, Any]):
        """Initializes the model with specific hyperparameters and sets up various components of SCARFLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SCARF. 
        """
        self.first_phase_loss = NTXentLoss(config["tau"])
        del config["tau"]
        del config["corruption_rate"]
        
        self._init_model(SCARF, config)
    
    def _get_first_phase_loss(self, batch):
        """Calculate the first phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        x, x_corrupted = batch
        emb_anchor, emb_corrupted = self.model(x, x_corrupted)

        loss = self.first_phase_loss(emb_anchor, emb_corrupted)

        return loss
    
    def _get_second_phase_loss(self, batch:Dict[str, Any]):
        """Calculate the second phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        x, y = batch
        y_hat = self(x).squeeze()

        loss = self.loss_fn(y_hat, y)
        
        return loss, y, y_hat
    
    def predict_step(self, batch, batch_idx: int
    ) -> torch.FloatTensor:
        """The perdict step of SCARF

        Args:
            batch (Dict[str, Any]): The input batch
            batch_idx (int): For compatibility, do not use

        Returns:
            torch.FloatTensor: The predicted output (logit)
        """
        
        y_hat = self(batch)

        return y_hat


from typing import Dict, Any, Type
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import VIME
from ts3l.utils.vime_utils import VIMEConfig

class VIMELightning(TS3LLightining):
    
    def __init__(self, config: VIMEConfig) -> None:
        """Initialize the pytorch lightining module of VIME

        Args:
            config (VIMEConfig): The configuration of VIMELightning.
        """
        super(VIMELightning, self).__init__(config)
    
    def _initialize(self, config: Dict[str, Any]) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of VIMELightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for VIME. 
        """
        
        self.alpha1 = config["alpha1"]
        self.alpha2 = config["alpha2"]
        del config["alpha1"]
        del config["alpha2"]
        
        self.beta = config["beta"]
        del config["beta"]
        
        self.K = config["K"]
        self.consistency_len = self.K + 1
        del config["K"]
        
        self.num_categoricals, self.num_continuous = config["num_categoricals"], config["num_continuous"]
        del config["num_categoricals"]
        del config["num_continuous"]
        
        self.u_label = config["u_label"]
        del config["u_label"]
        
        del config["p_m"]
        
        self.first_phase_mask_loss = nn.BCELoss()
        self.first_phase_feature_loss1 = nn.CrossEntropyLoss()
        self.first_phase_feature_loss2 = nn.MSELoss()
        
        self.consistency_loss = nn.MSELoss()

        self._init_model(VIME, config)
    
    def _get_first_phase_loss(self, batch:Dict[str, Any]):
        """Calculate the first phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        mask_output, feature_output = self.model(batch["input"])
        
        mask_loss = self.first_phase_mask_loss(mask_output, batch["label"][0])
        feature_loss1, feature_loss2 = 0, 0
        if self.num_categoricals > 0:
            feature_loss1 = self.first_phase_feature_loss1(feature_output[:, :self.num_categoricals], batch["label"][1][:, :self.num_categoricals])
        if self.num_continuous > 0:
            feature_loss2 = self.first_phase_feature_loss2(feature_output[:, self.num_categoricals:], batch["label"][1][:, self.num_categoricals:])
        final_loss = mask_loss + self.alpha1 * feature_loss1 + self.alpha2 * feature_loss2

        return final_loss
    
    def _get_second_phase_loss(self, batch:Dict[str, Any]):
        """Calculate the second phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        x = batch["input"]
        y = batch["label"]
        
        unsupervised_loss = 0
        unlabeled = x[y == self.u_label]
        
        if len(unlabeled) > 0:
            u_y_hat = self.model(unlabeled)
            target = u_y_hat[::self.consistency_len]
            target = target.repeat(1, self.K).reshape((-1, u_y_hat.shape[-1]))
            preds = torch.stack([u_y_hat[i, :] for i in range(len(u_y_hat)) if i % self.consistency_len != 0], dim = 0)
            unsupervised_loss += self.consistency_loss(preds, target)
        
        labeled_x = x[y != self.u_label]
        labeled_y = y[y != self.u_label]

        y_hat = self.model(labeled_x).squeeze()

        supervised_loss = self.loss_fn(y_hat, labeled_y)
        loss = supervised_loss + self.beta * unsupervised_loss
        
        return loss, labeled_y, y_hat
    
    def predict_step(self, batch, batch_idx: int
        ) -> torch.FloatTensor:
            """The predict step of VIME

            Args:
                batch (Dict[str, Any]): The input batch
                batch_idx (int): Only for compatibility, do not use

            Returns:
                torch.FloatTensor: The predicted output (logit)
            """
            y_hat = self(batch["input"])

            return y_hat


from typing import Dict, Any, Type, Tuple, Union
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import SwitchTab
from ts3l.utils.switchtab_utils import SwitchTabConfig

class SwitchTabLightning(TS3LLightining):
    
    def __init__(self, config: SwitchTabConfig) -> None:
        """Initialize the pytorch lightining module of SwitchTab

        Args:
            config (SwitchTabConfig): The configuration of SwitchTabLightning.
        """
        super(SwitchTabLightning, self).__init__(config)
    
    def _initialize(self, config: Dict[str, Any]) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of SwitchTabLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SwitchTab. 
        """
        
        self.u_label = config["u_label"]
        self.alpha = config["alpha"]
        del config["u_label"]
        del config["alpha"]
        del config["corruption_rate"]
        
        self.first_phase_reconstruction_loss = nn.MSELoss()
        
        self._init_model(SwitchTab, config)
    
    def _get_first_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        xs, xcs, ys = batch
        size = len(batch)
        
        xs = torch.concat([torch.concat([xs[:size], xs[:size]]), torch.concat([xs[size:], xs[size:]])])
        
        cls_idx = (ys != self.u_label)
        
        x_hat, y_hat = self.model(xcs)
        
        recon_loss = self.first_phase_reconstruction_loss(x_hat, xs)
        
        cls_loss = 0
        if sum(cls_idx) > 0:
            cls_loss = self.loss_fn(y_hat[cls_idx].squeeze(), ys[cls_idx])
        
        return recon_loss + self.alpha * cls_loss
    
    def _get_second_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Calculate the second phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]: The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        x, y = batch

        y_hat = self.model(x).squeeze()

        loss = self.loss_fn(y_hat, y)
        
        return loss, y, y_hat
    
    def predict_step(self, batch: torch.Tensor, batch_idx: int
        ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
            """The predict step of SwitchTab

            Args:
                batch (torch.Tensor): The input batch
                batch_idx (int): Only for compatibility, do not use

            Returns:
                Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]: The predicted output (logit) or (logit, salient_feature)
            """
            out = self(batch)

            return out
        
    def return_salient_feature(self, flag: bool) -> None:
        """Configures the model to either return or not return salient features based on the provided flag.

        Args:
            flag (bool): A boolean flag that determines the behavior of returning salient features.
                        If True, the model will include salient features in its output; if False, it will not.
        """
        self.model.return_salient_feature(flag)


from typing import Tuple

import torch
import torch.nn as nn

from .vime_self import VIMESelfSupervised
from .vime_semi import VIMESemiSupervised

class VIME(nn.Module):
    def __init__(self, 
                input_dim: int, hidden_dim: int, output_dim: int):
        """Initialize VIME

        Args:
            input_dim (int): The dimension of the encoder
            hidden_dim (int): The hidden dimension of the predictor
            output_dim (int): The output dimension of the predictor
        """
        super().__init__()
        
        self.self_net = VIMESelfSupervised(input_dim)
        self.semi_net = VIMESemiSupervised(input_dim, hidden_dim, output_dim)
        
        self.set_first_phase()
    
    def set_first_phase(self):
        """Set first phase step as the forward pass
        """
        self.forward = self.__first_phase_step
    
    def set_second_phase(self):
        """Set second phase step as the forward pass
        """
        self.forward = self.__second_phase_step
        
    def __first_phase_step(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """The first phase step of VIME

        Args:
            x (torch.Tensor): The input batch

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The predicted mask vector and predicted features
        """
        mask_output, feature_output = self.self_net(x)
        return mask_output, feature_output
    
    
    def __second_phase_step(self, x):
        """The second phase step of VIME

        Args:
            x (torch.FloatTensor): The input batch.

        Returns:
            torch.FloatTensor: The predicted logits of VIME
        """
        x = self.self_net.h(x)
        logits = self.semi_net(x)
        return logits


from typing import Dict, Any, Tuple, Union, Type

from .base_module import TS3LLightining
from ts3l.models import SubTab
from ts3l.utils.subtab_utils import JointLoss

import torch
from ts3l.utils.subtab_utils import SubTabConfig

class SubTabLightning(TS3LLightining):
    
    def __init__(self, config: SubTabConfig) -> None:
        """Initialize the pytorch lightining module of SubTab

        Args:
            config (SubTabConfig): The configuration of SubTabLightning.
        """
        super(SubTabLightning, self).__init__(config)

    def _initialize(self, config: Dict[str, Any]):
        """Initializes the model with specific hyperparameters and sets up various components of SubTabLightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for SubTab. 
        """
        self.first_phase_loss = JointLoss(
                                        config["tau"],
                                        config["n_subsets"],
                                        config["use_contrastive"],
                                        config["use_distance"],
                                        use_cosine_similarity = config["use_cosine_similarity"]
                                        )

        self.n_subsets = config["n_subsets"]

        del config["tau"],
        del config["use_contrastive"]
        del config["use_distance"]
        del config["use_cosine_similarity"]
        del config["shuffle"]
        del config["mask_ratio"]
        del config["noise_type"]
        del config["noise_level"]
        
        self._init_model(SubTab, config)
    
    def __get_recon_label(self, label: torch.Tensor) -> torch.Tensor:
        """Duplicates the input label tensor across the batch dimension to match the number of subsets for reconstruction loss.

        Args:
            label (torch.Tensor): The input tensor representing the label to be duplicated.

        Returns:
            torch.Tensor: A tensor with the input `label` duplicated `self.n_subsets` times along the batch dimension.
        """
        recon_label = label
        for _ in range(1, self.n_subsets):
            recon_label = torch.concat((recon_label, label), dim = 0)
        return recon_label
    
    def __arange_subsets(self, projections: torch.Tensor) -> torch.Tensor:
        """
        Rearranges the projections tensor into a sequence of subsets for first phase loss.
        This method takes a tensor of projections and reshapes it into a format where each subset of projections is concatenated along the first dimension. 
        The original tensor, which is assumed to represent a series of projections, is first reshaped into a tensor of shape `(n_subsets, samples, dim)` 
        where `n_subsets` is the number of subsets, `samples` is the batch_size, and `dim` is the embedding dimension. 
        The method then concatenates these subsets along the zeroth dimension to produce a sequence of projections suitable for first phase loss.

        Args:
            projections (torch.Tensor): A tensor of shape `(no, dim)` where `no` is the total number of observations and `dim` is the embedding dimension. 

        Returns:
            torch.Tensor: A reshaped tensor where subsets of projections are concatenated along the first dimension. 
                                The resulting shape is `(no, dim)`, maintaining the original dimensionality but rearranging the order of observations to align with subset divisions.
        """
        no, dim = projections.shape
        samples = int(no / self.n_subsets)
        
        projections = projections.reshape((self.n_subsets, samples, dim))
        return torch.concat([projections[:, i] for i in range(samples)])

    def _get_first_phase_loss(self, batch:Tuple[torch.FloatTensor, torch.Tensor, torch.Tensor]) -> torch.FloatTensor:
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.FloatTensor, torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        x, y_recons, y = batch
        projections, x_recons, x = self.model(x)
        
        recon_label = self.__get_recon_label(y_recons)
        
        projections = self.__arange_subsets(projections)
        
        total_loss, contrastive_loss, recon_loss, dist_loss = self.first_phase_loss(projections, x_recons, recon_label)

        return total_loss
    
    def _get_second_phase_loss(self, batch:Dict[str, Any]):
        """Calculate the second phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        x, _, y = batch
        y_hat = self(x).squeeze()
        
        loss = self.loss_fn(y_hat, y)
        
        return loss, y, y_hat
    
    def predict_step(self, batch, batch_idx: int) -> torch.FloatTensor:
        """The perdict step of SubTab

        Args:
            batch (Dict[str, Any]): The input batch
            batch_idx (int): For compatibility, do not use

        Returns:
            torch.FloatTensor: The predicted output (logit)
        """
        x, _, _ = batch
        y_hat = self(x)
        return y_hat


from typing import Dict, Any, Type, Tuple
import torch
from torch import nn

from .base_module import TS3LLightining
from ts3l.models import DAE
from ts3l.utils.dae_utils import DAEConfig

class DAELightning(TS3LLightining):
    
    def __init__(self, config: DAEConfig) -> None:
        """Initialize the pytorch lightining module of DAE

        Args:
            config (DAEConfig): The configuration of DAELightning.
        """
        super(DAELightning, self).__init__(config)
    
    def _initialize(self, config: Dict[str, Any]) -> None:
        """Initializes the model with specific hyperparameters and sets up various components of DAELightning.

        Args:
            config (Dict[str, Any]): The given hyperparameter set for DAE. 
        """
        
        self.mask_loss_weight = config["mask_loss_weight"]
        del config["mask_loss_weight"]
        
        self.num_categoricals, self.num_continuous = config["num_categoricals"], config["num_continuous"]
        del config["num_categoricals"]
        del config["num_continuous"]
        
        del config["noise_type"]
        del config["noise_level"]
        del config["noise_ratio"]
        
        self.first_phase_mask_loss = nn.BCELoss()
        self.first_phase_feature_loss1 = nn.CrossEntropyLoss()
        self.first_phase_feature_loss2 = nn.MSELoss()

        self._init_model(DAE, config)
    
    def _get_first_phase_loss(self, batch: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]):
        """Calculate the first phase loss

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): The input batch

        Returns:
            torch.FloatTensor: The final loss of first phase step
        """
        x, x_bar, mask = batch
        mask_output, feature_output = self.model(x_bar)
        
        # print(mask, mask_output)
        mask_loss = self.first_phase_mask_loss(mask_output, mask)
        feature_loss1, feature_loss2 = 0, 0
        if self.num_categoricals > 0:
            feature_loss1 = self.first_phase_feature_loss1(feature_output[:, :self.num_categoricals], x[:, :self.num_categoricals])
        if self.num_continuous > 0:
            feature_loss2 = self.first_phase_feature_loss2(feature_output[:, self.num_categoricals:], x[:, self.num_categoricals:])
        final_loss = mask_loss * self.mask_loss_weight + feature_loss1 + feature_loss2

        return final_loss
    
    def _get_second_phase_loss(self, batch:Dict[str, Any]):
        """Calculate the second phase loss

        Args:
            batch (Dict[str, Any]): The input batch

        Returns:
            torch.FloatTensor: The final loss of second phase step
            torch.Tensor: The label of the labeled data
            torch.Tensor: The predicted label of the labeled data
        """
        x, y = batch

        y_hat = self.model(x).squeeze()

        loss = self.loss_fn(y_hat, y)
        
        return loss, y, y_hat
    
    def predict_step(self, batch, batch_idx: int
        ) -> torch.FloatTensor:
            """The predict step of DAE

            Args:
                batch (torch.Tensor): The input batch
                batch_idx (int): Only for compatibility, do not use

            Returns:
                torch.FloatTensor: The predicted output (logit)
            """
            y_hat = self(batch)

            return y_hat


import pandas as pd
from typing import Dict, Any, List, Union, Optional, Tuple
from numpy.typing import NDArray

import torch
from torch.utils.data import Dataset

import numpy as np
from dataclasses import asdict
from ts3l.utils.vime_utils import VIMEConfig

class VIMEDataset(Dataset):
    def __init__(self,
                X: pd.DataFrame,
                Y: Optional[Union[NDArray[np.int_], NDArray[np.float_]]] = None, 
                config: Optional[VIMEConfig] = None, 
                unlabeled_data: Optional[pd.DataFrame] = None, 
                continuous_cols: Optional[List] = None, 
                category_cols: Optional[List] = None, 
                u_label: Any = -1, 
                is_second_phase: Optional[bool] = False,
                is_regression: Optional[bool] = False, 
        ) -> None:
        """A dataset class for VIME that handles labeled and unlabeled data.

        This class is designed to manage data for the VIME, accommodating both labeled and unlabeled datasets
        for self- and semi-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Union[NDArray[np.int_], NDArray[np.float_]], optional): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            config (Dict[str, Any]): The given hyperparameter set for VIME.
            unlabeled_data (pd.DataFrame): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            continuous_cols (List, optional): List of continuous columns. Defaults to None.
            category_cols (List, optional): List of categorical columns. Defaults to None.
            u_label (int, optional): The specifier for unlabeled sample. Defaults to -1.
            is_second_phase (bool): The flag that determines whether the dataset is for first phase or second phase learning. Default is False.
            is_regression (bool, optional): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
        """
        
        self.label = None
        
        if not is_second_phase:
            self.__getitem = self.__first_phase_get_item
            
        else:
            self.__getitem = self.__second_phase_get_item
            
            self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
            
            if Y is not None:
            
                self.label = self.label_class(Y)
            
                if unlabeled_data is not None:
                    self.label = torch.concat((self.label, self.label_class([u_label for _ in range(len(unlabeled_data))])), dim=0)
                
                if not is_regression:
                    class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                    num_samples = len(self.label)

                    class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                    self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]
                else:
                    self.weights = [1.0 for _ in range(len(X))]
                    if unlabeled_data is not None:
                        unlabeled_weight = len(X) / len(unlabeled_data)
                        self.weights.extend([unlabeled_weight for _ in range(len(unlabeled_data))])
        
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
            
        self.cont_data = torch.FloatTensor(X[continuous_cols].values)
        self.cat_data = torch.FloatTensor(X[category_cols].values)
        
        self.continuous_cols = continuous_cols
        self.category_cols = category_cols

        if config is not None:
            self.config = config
        
        self.u_label = u_label
    
    def __getitem__(self, idx: int) -> Dict[str, Union[torch.Tensor, Tuple[torch.Tensor, ...]]]:
        """Retrieves the feature and label tensors for a given index. The structure of the label 
            varies depending on the learning phase: 

            - In the first phase, the label is a tuple containing a mask and the original feature tensor.
            - In the second phase, the label is either a token indicating unlabeled samples or the actual label for labeled samples.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            Dict[Any, torch.Tensor]: The feature tensor and the label for the given index. 
            For first phase learning, label is a tuple of mask and original feature tensor.
            For second phase learning, label is a one of a token for unlabeled samples or a label of it.
        """
        return self.__getitem(idx)
    
    def __mask_generator(self, p_m, x):
        """Generate mask vector.
        
        Args:
            - p_m: corruption probability
            - x: feature matrix
            
        Returns:
            - mask: binary mask matrix 
        """
        mask = np.random.binomial(1, p_m, x.shape)

        return np.expand_dims(mask, axis=0)

    def __pretext_generator(self, m, x, empirical_dist) -> Tuple[torch.LongTensor, torch.FloatTensor]:  
        """Generate corrupted samples.
        
        Args:
            m: mask matrix
            x: feature matrix
            
        Returns:
            m_new: final mask matrix after corruption
            x_tilde: corrupted feature matrix
        """
        
        # Parameters
        dim = x.shape[0]
        # Randomly (and column-wise) shuffle data
        x_bar = np.zeros([1, dim])

        rand_idx = np.random.randint(0, len(empirical_dist), size=dim)
        
        x_bar = np.array([empirical_dist[rand_idx[i], i] for i in range(dim)])
        
        # Corrupt samples
        x_tilde = x * (1-m) + x_bar * m  
        # Define new mask matrix
        m_new = 1 * (x != x_tilde)
        
        if dim != 1:
            return m_new.squeeze(), x_tilde.squeeze()
        else:
            return m_new.squeeze().reshape(1), x_tilde.squeeze().reshape(1)
    
    def __generate_x_tildes(self, cat_samples: torch.Tensor, cont_samples:torch.Tensor) -> torch.Tensor:
        """Generate x_tilde for consistency regularization

        Args:
            cat_samples (torch.Tensor): The categorical features to generate x_tilde
            cont_samples (torch.Tensor): The continuous features to generate x_tilde

        Returns:
            torch.Tensor: x_tilde for consistency regularization
        """
        m_unlab = self.__mask_generator(self.config.p_m, cat_samples)
        dcat_m_label, cat_x_tilde = self.__pretext_generator(m_unlab, cat_samples, self.cat_data)
        
        m_unlab = self.__mask_generator(self.config.p_m, cont_samples)
        cont_m_label, cont_x_tilde = self.__pretext_generator(m_unlab, cont_samples, self.cont_data)
        x_tilde = torch.concat((cat_x_tilde, cont_x_tilde)).float()
        
        return x_tilde
    
    def __first_phase_get_item(self, idx: int) -> Dict[str, Union[torch.Tensor, Tuple[torch.Tensor, ...]]]:
        """Return a input and label pair

        Args:
            idx (int): The index of the data to sample

        Returns:
            Dict[str, Union[torch.Tensor, Tuple[torch.Tensor, ...]]]: A pair of input and label for first phase learning
        """
        cat_samples = self.cat_data[idx]
        m_unlab = self.__mask_generator(self.config.p_m, cat_samples)
        cat_m_label, cat_x_tilde = self.__pretext_generator(m_unlab, cat_samples, self.cat_data)

        cont_samples = self.cont_data[idx]
        m_unlab = self.__mask_generator(self.config.p_m, cont_samples)
        cont_m_label, cont_x_tilde = self.__pretext_generator(m_unlab, cont_samples, self.cont_data)

        m_label = torch.concat((cat_m_label, cont_m_label)).float()
        x_tilde = torch.concat((cat_x_tilde, cont_x_tilde)).float()

        x = torch.concat((cat_samples, cont_samples))

        return {
                "input" : x_tilde,
                "label" : (m_label, x)
                }

    def __second_phase_get_item(self, idx) -> Dict[str, Union[torch.Tensor, Tuple[torch.Tensor, ...]]]:
        """Return a input and label pair

        Args:
            idx (int): The index of the data to sample

        Returns:
            Dict[str, Union[torch.Tensor, Tuple[torch.Tensor, ...]]]: A pair of input and label for second phase learning
        """
        cat_samples = self.cat_data[idx]
        cont_samples = self.cont_data[idx]
        x = torch.concat((cat_samples, cont_samples)).squeeze()
        if self.label is not None:
            
            if self.label[idx] == self.u_label:
                _xs = [x]
                _xs.extend([self.__generate_x_tildes(cat_samples, cont_samples) for _ in range(self.config.K)])
                xs = torch.stack(_xs)
                
                return {
                    "input" : xs,
                    "label" : self.label_class([self.u_label for _ in range(len(xs))])
                }
            else:
                return {
                    "input" : x.unsqueeze(0),
                    "label" : self.label[idx].unsqueeze(0)
                }
        else:
            return {
                    "input" : x,
                    "label" : self.u_label
            }
            
    def __len__(self):
        """Return the length of the dataset
        """
        return len(self.cat_data)

class VIMESemiSLCollateFN(object):
    """A callable class designed for batch processing, specifically tailored for the second phase learning with VIME. 
    It consolidates a batch of samples into a single dictionary with concatenated inputs and labels, suitable for model input.

    This class is meant to be used as a collate function in a DataLoader, where it efficiently organizes batch data 
    for training during second phase learning.
    """
    def __call__(self, batch):
        return {
            'input': torch.concat([x['input'] for x in batch], dim=0),
            'label': torch.concat([x['label'] for x in batch], dim=0)
        }


import torch
from torch import nn
from torch.nn import functional as F

class NTXentLoss(nn.Module):
    def __init__(self, temperature=1.0):
        """NT-Xent loss for contrastive learning using cosine distance as similarity metric as used in [SimCLR](https://arxiv.org/abs/2002.05709).
        Implementation adapted from https://theaisummer.com/simclr/#simclr-loss-implementation
        Args:
            temperature (float, optional): scaling factor of the similarity metric. Defaults to 1.0.
        """
        super().__init__()
        self.temperature = temperature

    def forward(self, z_i, z_j):
        """Compute NT-Xent loss using only anchor and positive batches of samples. Negative samples are the 2*(N-1) samples in the batch
        Args:
            z_i (torch.tensor): anchor batch of samples
            z_j (torch.tensor): positive batch of samples
        Returns:
            float: loss
        """
        batch_size = z_i.size(0)

        # compute similarity between the sample's embedding and its corrupted view
        z = torch.cat([z_i, z_j], dim=0)
        similarity = F.cosine_similarity(z.unsqueeze(1), z.unsqueeze(0), dim=2)

        sim_ij = torch.diag(similarity, batch_size)
        sim_ji = torch.diag(similarity, -batch_size)
        positives = torch.cat([sim_ij, sim_ji], dim=0)

        mask = (~torch.eye(batch_size * 2, batch_size * 2, dtype=torch.bool)).float()
        numerator = torch.exp(positives / self.temperature).to(mask.device)
        denominator = mask * torch.exp(similarity / self.temperature).to(mask.device)

        all_losses = -torch.log(numerator / torch.sum(denominator, dim=1))
        loss = torch.sum(all_losses) / (2 * batch_size)

        return loss


import torch
from torch import nn
import numpy as np
import itertools

class JointLoss(nn.Module):
    """ JointLoss for SubTab during the first phase learning.
    
    When computing the constrastive loss, we use a similarity matrix of size (N x k) x N. The matrix includes k positive samples and all other samples are considered negatives. 
    The matrix is shown below as an 8x4 array, assuming a batch size of 4 and 2 subsets.
                                                        P . . .
                                                        P . . .
                                                        . P . .
                                                        . P . .
                                                        . . P .
                                                        . . P .
                                                        . . . P
                                                        . . . P
    """

    def __init__(self,
                 # batch_size: int,
                 tau: float,
                 n_subsets: int,
                 use_contrastive: bool = True,
                 use_distance: bool = True,
                 use_cosine_similarity: bool = False
        ) -> None:
        super(JointLoss, self).__init__()

        # n_subsets
        self.n_subsets = n_subsets
        # Temperature to use scale logits
        self.temperature = tau

        # Function to generate similarity matrix: Cosine, or Dot product
        self.similarity_fn = self._cosine_simililarity if use_cosine_similarity else self._dot_simililarity
        # Loss function
        self.mse_loss = nn.MSELoss()
        self.criterion = nn.CrossEntropyLoss(reduction="sum")
        
        self.use_contrastive = use_contrastive
        self.use_distance = use_distance
        

    @staticmethod
    def _dot_simililarity(x, y):
        # Reshape x: (2N, C) -> (2N, 1, C)
        x = x.unsqueeze(1)
        # Reshape y: (2N, C) -> (1, C, 2N)
        y = y.T.unsqueeze(0)
        # Similarity shape: (2N, 2N)
        similarity = torch.tensordot(x, y, dims=2)
        return similarity

    def _cosine_simililarity(self, x, y):
        similarity = nn.CosineSimilarity(dim=-1)
        # Reshape x: (2N, C) -> (2N, 1, C)
        x = x.unsqueeze(1)
        # Reshape y: (2N, C) -> (1, C, 2N)
        y = y.unsqueeze(0)
        # Similarity shape: (2N, 2N)
        return similarity(x, y)
    
    def get_anchor_loss(self, turn, similarity):
        group_start = (turn // self.n_subsets) * self.n_subsets
        group_end = min(group_start + self.n_subsets, len(similarity))

        # Select the positive and negative pairs for the anchor
        positives = similarity[group_start:group_end]
        negatives = torch.cat((similarity[:group_start], similarity[group_end:]), dim=0)

        # Exclude self-similarity
        positives = positives[positives != 1]

        # Loss calculation for the anchor
        pos_sum = torch.sum(torch.exp(positives))
        neg_sum = torch.sum(torch.exp(negatives))
        # return -torch.log(pos_sum / (pos_sum + neg_sum))
        anchor_loss = -torch.log(pos_sum / (pos_sum + neg_sum))
        return anchor_loss
        
    def XNegloss(self, projections: torch.FloatTensor) -> torch.Tensor:
        
        
        # Compute cosine similarity using the provided function
        similarity = self.similarity_fn(projections, projections) / self.temperature
        
        # return torch.stack([self.get_anchor_loss(turn, similarity[turn]) for turn in range(len(similarity))]).mean()
        anchor_losses = [self.get_anchor_loss(turn, similarity[turn]) for turn in range(len(similarity))]
        
        return torch.stack(anchor_losses).mean()

    def forward(self, projections, xrecon, xorig):
        """
        Args:
            projections (torch.FloatTensor): Projections for each subset.
            xrecon (torch.FloatTensor): Reconstructed sample x from subsets.
            xorig (torch.FloatTensor): Original features of x
        """
        
        # recontruction loss
        recon_loss = self.mse_loss(xrecon, xorig)

        # Initialize contrastive and distance losses with recon_loss as placeholder
        closs, dist_loss = None, None

        # Start with default loss i.e. reconstruction loss
        loss = recon_loss

        if self.use_contrastive:
            closs = self.XNegloss(projections)
            loss = loss + closs

        if self.use_distance:
            # recontruction loss for z
            combi = np.array(list(itertools.combinations(range(self.n_subsets), 2)))
            left = combi[:, 0]
            right = combi[:, 1]
            dist_loss = torch.stack([self.mse_loss(projections[i:i + self.n_subsets][left], projections[i:i + self.n_subsets][right]) for i in range(0, len(projections), self.n_subsets)]).mean()
            loss += dist_loss

        return loss, closs, recon_loss, dist_loss



from .datamodule import TS3LDataModule
from .misc import RegressionMetric, ClassificationMetric, get_category_dims
from .base_config import BaseConfig

__all__ = ["TS3LDataModule", "vime_utils", "subtab_utils", "scarf_utils", "RegressionMetric", "ClassificationMetric", "BaseConfig", "get_category_dims"]


import torch
import torch.nn as nn
from torch.nn import functional as F

import numpy as np
from numpy.typing import NDArray

import itertools

from types import SimpleNamespace
from typing import Dict, Any, Tuple, List, Union

class ShallowEncoder(nn.Module):
    def __init__(self,
                 feat_dim : int,
                 hidden_dim : int,
                 n_subsets : int,
                 overlap_ratio : float,
    ) -> None:
        super().__init__()

        n_column_subset = int(feat_dim / n_subsets)
        n_overlap = int(overlap_ratio * n_column_subset)

        self.net = nn.Sequential(
            nn.Linear(n_column_subset + n_overlap, hidden_dim),
            nn.LeakyReLU(),
        )
        
    def forward(self,
                x : torch.Tensor
    ) -> torch.Tensor:
        return self.net(x)

class ShallowDecoder(nn.Module):
    def __init__(self,
                 hidden_dim : int,
                 out_dim : int
    ) -> None:
        super().__init__()

        self.net = nn.Linear(hidden_dim, out_dim)
    
    def forward(self, x : torch.Tensor) -> torch.Tensor:
        return self.net(x)

class AutoEncoder(nn.Module):
    def __init__(self,
                 feat_dim : int,
                 hidden_dim : int,
                 n_subsets : int,
                 overlap_ratio : float,
    ) -> None:
        super().__init__()

        self.encoder = ShallowEncoder(feat_dim, hidden_dim, n_subsets, overlap_ratio)
        self.decoder = ShallowDecoder(hidden_dim, feat_dim)

        self.projection_net = nn.Sequential(
            nn.Linear(hidden_dim, hidden_dim),
            nn.LeakyReLU(),
            nn.Linear(hidden_dim, hidden_dim),
        )
    
    def encode(self, x : torch.Tensor) -> torch.Tensor:
        return self.encoder(x)
    
    def decode(self, x : torch.Tensor) -> torch.Tensor:
        return self.decoder(x)
    
    def forward(self, x : torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        latent = self.encode(x)
        projection = self.projection_net(latent)
        projection = F.normalize(projection, p = 2, dim = 1)
        x_recon = self.decode(latent)
        return latent, projection, x_recon

class SubTab(nn.Module):

    def __init__(self,
                 input_dim: int,
                 output_dim: int,
                 hidden_dim: int,
                 
                 n_subsets: int,
                 overlap_ratio: float,
    ) -> None:
        super().__init__()

        self.feat_dim = input_dim
        
        self.n_subsets = n_subsets
        
        self.ae = AutoEncoder(self.feat_dim, hidden_dim, n_subsets, overlap_ratio)
        
        self.head = nn.Sequential(
            nn.Linear(hidden_dim, output_dim)
        )
        self.set_first_phase()
        
    

    def set_first_phase(self) -> None:
        self.forward = self.__first_phase_step
    
    def set_second_phase(self) -> None:
        self.forward = self.__second_phase_step

    def __first_phase_step(self, x : torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

        latents, projections, x_recons = self.ae(x)
        
        return projections, x_recons, x
    
    def __arange_subsets(self, latent: torch.Tensor) -> torch.Tensor:
        no, dim = latent.shape
        samples = int(no / self.n_subsets)
        latent = latent.reshape((self.n_subsets, samples, dim))
        return torch.concat([latent[:, i] for i in range(samples)])
    
    def __second_phase_step(self, 
                  x : torch.Tensor,
                  return_embeddings : bool = False) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:

        latent = self.ae.encode(x)
        latent = self.__arange_subsets(latent)
        
        latent = latent.reshape(x.shape[0] // self.n_subsets, self.n_subsets, -1).mean(1)
        out = self.head(latent)

        if return_embeddings:
            return out, latent
        return out




from typing import Tuple, List, Union

import torch
from torch import nn
from torch.nn import functional as F

class FeatureTokenizer(nn.Module):

    def __init__(self, 
                emb_dim : int,
                cont_nums : int,
                cat_dims : List[int]
        ) -> None:
        super().__init__()
        self.emb_dim = emb_dim
        self.cont_nums = cont_nums
        self.cat_nums = len(cat_dims)
        
        self.cat_dims = cat_dims

        bias_dim = 0

        if cont_nums is not None:
            bias_dim += cont_nums

        if cat_dims is not None:
            bias_dim += len(cat_dims)

            category_offsets = torch.tensor([0] + cat_dims[:-1]).cumsum(0)
            self.register_buffer('category_offsets', category_offsets)

            self.cat_weights = nn.Embedding(sum(cat_dims), emb_dim)
        
        self.weight = nn.Parameter(torch.Tensor(cont_nums + 1, emb_dim))
        self.bias = nn.Parameter(torch.Tensor(bias_dim, emb_dim))

    def forward(self, 
                x: torch.Tensor
        ) -> torch.Tensor:

        x_cats, x_conts = x[:, :self.cat_nums].long(), x[:, self.cat_nums:]
        
        x_conts = torch.cat(
            [torch.ones(
                len(x_conts) if x_conts is not None else len(x_cats), 1, 
                device = x_conts.device if x_conts is not None else x_cats.device)
            ]
            + ([] if x_conts is None else [x_conts]),
            dim = 1
        )

        x = self.weight[None] * x_conts[:, :, None]

        if x_cats is not None:

            x = torch.cat(
                [x, self.cat_weights(x_cats + self.category_offsets[None])],
                dim=1,
            )
        if self.bias is not None:
            bias = torch.cat(
                [
                    torch.zeros(1, self.bias.shape[1], device=x.device),
                    self.bias,
                ]
            )
            x = x + bias[None]

        return x

class MultiheadAttention(nn.Module):
    def __init__(self,
                 emb_dim : int,
                 n_heads : int,
                 attn_dropout : float,
    ) -> None:
        super().__init__()
        
        assert emb_dim % n_heads == 0

        self.emb_dim = emb_dim
        self.n_heads = n_heads
        self.head_dim = emb_dim // n_heads

        self.W_qkv = nn.Linear(emb_dim, emb_dim * 3)

        self.W_out = nn.Linear(emb_dim, emb_dim)
        
        self.scale = emb_dim ** - 0.5

        self.dropout = nn.Dropout(attn_dropout) if attn_dropout > 0 else None

    
    def forward(self,
                x: torch.Tensor,
                return_attn: bool = False
    ) -> Union[Tuple[torch.Tensor, torch.Tensor], torch.Tensor]:
        
        batch_size = len(x)

        
        qkv = self.W_qkv(x)
        qkv = qkv.reshape(batch_size, -1, self.n_heads, self.head_dim * 3)
        qkv = qkv.permute(0, 2, 1, 3)
        q, k, v = qkv.chunk(3, dim = -1)
        
        attention = F.softmax(q @ k.transpose(-2, -1) / self.scale, dim=-1)
        attention = F.softmax(attention, dim = -1)
        if self.dropout is not None:
            attention = self.dropout(attention)
        v = attention @ v

        v = v.permute(0, 2, 1, 3)
        v = v.reshape(batch_size, -1, self.emb_dim)

        out = self.W_out(v)

        if return_attn:
            return out, attention
        else:
            return out
        
def reglu(x: torch.Tensor) -> torch.Tensor:
    a, b = x.chunk(2, dim=-1)
    return a * F.relu(b)

class Transformer(nn.Module):

    def __init__(self,
                 emb_dim: int, 
                 n_heads: int, 
                 attn_dropout: float,
                 ffn_dropout: float,
                 ffn_factor_dim: float,
                 layer_idx: int,
    ) -> None:
        super().__init__()

        hidden_dim = int(emb_dim * ffn_factor_dim)

        self.block = nn.ModuleDict(
            {
                'attention' : MultiheadAttention(emb_dim, n_heads, attn_dropout),
                'linear0' : nn.Linear(emb_dim, hidden_dim * 2),
                'linear1' : nn.Linear(hidden_dim, emb_dim),
                'norm1': nn.LayerNorm(emb_dim)
            }
        )
        if layer_idx:
            self.block['norm0'] = nn.LayerNorm(emb_dim)
        if ffn_dropout > 0:
            self.block['ffn_dropout'] = nn.Dropout(ffn_dropout)

        self.activation = reglu

    def forward(self,
                x: torch.Tensor
    ) -> torch.Tensor:

        if 'norm0' in self.block:
            x_res = self.block['attention'](self.block['norm0'](x))
        else:
            x_res = self.block['attention'](x)
        
        x = x + x_res

        x_res = self.block['norm1'](x)
        x_res = self.block['linear0'](x_res)
        x_res = self.activation(x_res)
        x_res = self.block['ffn_dropout'](x_res)
        x_res = self.block['linear1'](x_res)

        x = x + x_res

        return x


class FTTransformer(nn.Module):

    def __init__(self,
                 cont_nums: int,
                 cat_dims: List[int],
                 emb_dim: int, 
                 n_heads: int, 
                 attn_dropout: float,
                 ffn_dropout: float,
                 ffn_factor_dim: float,
                 depth: int,
    ) -> None:
        super().__init__()

        self.tokenizer = FeatureTokenizer(emb_dim, cont_nums, cat_dims)

        self.backbone = nn.Sequential(
            *[
                Transformer(
                    emb_dim,
                    n_heads, 
                    attn_dropout, 
                    ffn_dropout, 
                    ffn_factor_dim,
                    idx
                )
                for idx in range(depth)
            ],
        )
            
            
    def forward(self,
                x: torch.Tensor
    ) -> torch.Tensor:

        x = self.tokenizer(x)

        x = self.backbone(x)

        x = x[:, 0]

        return x


from numpy.typing import NDArray
from typing import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.uniform import Uniform

import numpy as np

from ts3l.models.common import MLP


class DAE(nn.Module):
    def __init__(
        self,
        input_dim,
        hidden_dim,
        encoder_depth=4,
        head_depth=2,
        dropout_rate = 0.04,
        output_dim = 2,
    ):
        """Implementation of Denoising AutoEncoder.
        DAE processes input data that has been partially corrupted, producing clean data during the self-supervised learning stage. 
        The denoising task enables the model to learn the input distribution and generate latent representations that are robust to corruption. 
        These latent representations can be utilized for a variety of downstream tasks.
        Args:
            input_dim (int): The size of the inputs
            hidden_dim (int): The dimension of the hidden layers
            encoder_depth (int, optional): The number of layers of the encoder MLP. Defaults to 4.
            head_depth (int, optional): The number of layers of the pretraining head. Defaults to 2.
            dropout_rate (float, optional): The probability of setting the outputs of the dropout layer to zero during training. Defaults to 0.04.
            output_dim (int, 2): The size of the outputs
        """
        super().__init__()

        self.encoder = MLP(input_dim, hidden_dim, encoder_depth, dropout_rate)
        self.mask_predictor_head = MLP(hidden_dim, input_dim, head_depth, dropout_rate)
        self.reconstruction_head = MLP(hidden_dim, input_dim, head_depth, dropout_rate)

        self.head = nn.Sequential(
            OrderedDict([
                ("head_activation", nn.ReLU(inplace=True)),
                ("head_batchnorm", nn.BatchNorm1d(hidden_dim)),
                ("head_dropout", nn.Dropout(dropout_rate)),
                ("head_linear", nn.Linear(hidden_dim, output_dim))
            ])
        )
        
    def set_first_phase(self):
        self.forward = self.__first_phase_step
    
    def set_second_phase(self):
        self.forward = self.__second_phase_step

    def __first_phase_step(self, x):

        emb = self.encoder(x)
        mask = torch.sigmoid(self.mask_predictor_head(emb))
        feature = self.reconstruction_head(emb)

        return mask, feature
    
    def __second_phase_step(self, x):
        emb = self.encoder(x)
        output = self.head(emb)
        return output



from typing import Dict, Any, Tuple, Union, Optional
from numpy.typing import NDArray

import torch
from torch.utils.data import Dataset
import numpy as np
import pandas as pd

from dataclasses import asdict
from ts3l.utils.subtab_utils import SubTabConfig
class SubTabDataset(Dataset):
    def __init__(self, X: pd.DataFrame, 
                        Y: Optional[Union[NDArray[np.int_], NDArray[np.float_]]] = None,
                        unlabeled_data: Optional[pd.DataFrame] = None, 
                        is_regression: Optional[bool] = False
                        ) -> None:
        """A dataset class for SubTab that handles labeled and unlabeled data.

        This class is designed to manage data for the SubTab, accommodating both labeled and unlabeled datasets
        for self-supervised learning scenarios. It supports regression and classification tasks.

        Args:
            X (pd.DataFrame): DataFrame containing the features of the labeled data.
            Y (Optional[Union[NDArray[np.int_], NDArray[np.float_]]]): Numpy array containing the labels for the data. 
                Use integers for classification labels and floats for regression targets. Defaults to None.
            unlabeled_data (Optional[pd.DataFrame]): DataFrame containing the features of the unlabeled data, used for 
                self-supervised learning. Defaults to None.
            is_regression (Optional[bool]): Flag indicating whether the task is regression (True) or classification (False).
                Defaults to False.
        """
        if unlabeled_data is not None:
            X = pd.concat([X, unlabeled_data])
            
        self.data = torch.FloatTensor(X.values)
        
        self.label_class = torch.FloatTensor if is_regression else torch.LongTensor
            
        if Y is None:
            self.label = None
        else:
            self.label = self.label_class(Y)
            
            if not is_regression:
                class_counts = [sum((self.label == i)) for i in set(self.label.numpy())]
                num_samples = len(self.label)

                class_weights = [num_samples/class_counts[i] for i in range(len(class_counts))]
                self.weights = [class_weights[self.label[i]] for i in range(int(num_samples))]
    
    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """Retrieves the feature and label tensors for a given index.

        Args:
            idx (int): The index of the data point to retrieve.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: The feature tensor and the label 
            tensor (if available) for the given index. If labels are not provided, a placeholder is returned.
        """
        if self.label is None:
            return self.data[idx], torch.LongTensor([0])
        return self.data[idx], self.label[idx]
    
    def __len__(self) -> int:
        """Returns the total number of items in the dataset.

        Returns:
            int: The size of the dataset.
        """
        return len(self.data)
    
class SubTabCollateFN(object):
    def __init__(self, config: SubTabConfig) -> None:
        """A collate function for SubTab to generate subsets of features for dataloaders.

        This collate function is designed to prepare batches for the SubTab model by generating subsets of features with
        optional noise and masking. It supports configurable shuffling, overlap, and masking of features.

        Args:
            config (SubTabConfig): The given hyperparameter set for SubTab.
        """
        
        self.config = config
            
        self.shuffle = self.config.shuffle
        
        self.n_subsets = self.config.n_subsets
        self.overlap_ratio = self.config.overlap_ratio
        self.mask_ratio = self.config.mask_ratio
        self.noise_type = self.config.noise_type
        self.noise_level = self.config.noise_level
        
        self.input_dim = self.config.input_dim
        self.n_column_subset = int(self.input_dim / self.n_subsets)
        # Number of overlapping features between subsets
        self.n_overlap = int(self.overlap_ratio * self.n_column_subset)
        self.column_idx = np.array(range(self.input_dim))
    
    def __generate_noisy_xbar(self, x : torch.Tensor) -> torch.Tensor:
        """Generates a noisy version of the input sample `x`.

        Args:
            x (torch.Tensor): The original sample.

        Returns:
            torch.Tensor: The noisy sample.
        """
        no, dim = x.shape
        
        # Initialize corruption array
        x_bar = torch.zeros([no, dim]).to(x.device)

        # Randomly (and column-wise) shuffle data
        if self.noise_type == "Swap":
            x_bar = torch.stack([x[np.random.permutation(no), i] for i in range(dim)], dim=1).to(x.device)
        elif self.noise_type == "Gaussian":
            x_bar = x + torch.normal(torch.zeros(x.shape), torch.full(x.shape, self.noise_level)).to(x.device)

        return x_bar
    
    def __generate_x_tilde(self, x: torch.Tensor, subset_column_idx: NDArray[np.int32]) -> torch.Tensor:
        """Generates a masked and potentially noisy subset of the input sample `x` based on the provided column indices.

        Args:
            x (torch.Tensor): The original sample.
            subset_column_idx (NDArray[np.int32]): Indices of columns to include in the subset.

        Returns:
            torch.Tensor: The processed subset of the sample.
        """
        x_bar = x[:, subset_column_idx]
        x_bar_noisy = self.__generate_noisy_xbar(x_bar)
        
        mask = torch.distributions.binomial.Binomial(total_count = 1, probs = self.mask_ratio).sample(x_bar.shape).to(x_bar.device)
        
        x_bar = x_bar * (1 - mask) + x_bar_noisy * mask
        
        return x_bar
    
    def __generate_subset(self,
                        x : torch.Tensor,
    ) -> torch.Tensor:
        """Generates subsets of features from the input sample `x`, applying shuffling, noise, and masking as configured.

        Args:
            x (torch.Tensor): The batch of samples.

        Returns:
            torch.Tensor: A tensor containing the generated subsets for the batch.
        """

        permuted_order = np.random.permutation(self.n_subsets) if self.shuffle else np.arange(self.n_subsets)

        subset_column_indice_list = [self.column_idx[:self.n_column_subset + self.n_overlap]]
        subset_column_indice_list.extend([self.column_idx[range(i * self.n_column_subset - self.n_overlap, (i + 1) * self.n_column_subset)] for i in range(self.n_subsets)])
        
        
        subset_column_indice = np.array(subset_column_indice_list)
        subset_column_indice = subset_column_indice[permuted_order]
        
        if len(subset_column_indice) == 1:
            subset_column_indice = np.concatenate([subset_column_indice, subset_column_indice])
        
        x_tildes = torch.concat([self.__generate_x_tilde(x, subset_column_indice[i]) for i in range(self.n_subsets)]) # [subset1, subset2, ... ,  subsetN]

        return x_tildes
    
    def __call__(self, batch: Tuple[torch.Tensor, torch.Tensor]) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Processes a batch of samples, and generating feature subsets.

        Args:
            batch (Tuple[torch.Tensor, torch.Tensor]): A batch of samples.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]: A tuple containing the processed feature subsets, the original samples, and the labels.
        """
        y_recons = torch.stack([sample[0] for sample in batch])
        x = self.__generate_subset(y_recons)
        y = torch.tensor([sample[1] for sample in batch])
        
        return x, y_recons, y